@@ -739,6 +739,62 @@ class gdaAttack:
         conn.close()
         return ret
 
+    def getAnonTableCharacteristics(self, tableName=''):
+        """Returns the full contents of the table characteristics
+
+           Return value is a dict indexed by column name: <br/>
+
+               { '<colName>':
+                   {
+                       'av_rows_per_vals': 3.93149,
+                       'av_uids_per_val': 0.468698,
+                       'column_label': 'continuous',
+                       'column_name': 'dropoff_latitude',
+                       'column_type': 'real',
+                       'max': '898.29382000000000',
+                       'min': '-0.56333297000000',
+                       'num_distinct_vals': 24216,
+                       'num_rows': 95205,
+                       'num_uids': 11350,
+                       'std_rows_per_val': 10.8547,
+                       'std_uids_per_val': 4.09688},
+                   }
+               }
+
+        """
+        if len(tableName) == 0:
+            # caller didn't supply a table name, so get it from the
+            # class init
+            tableName = self._p['table']
+
+        # Modify table name to the default for the characteristics table
+        # tableName += '_char'
+
+        # Establish connection to database
+        db = getDatabaseInfo(self._p['anonDb'])
+        connStr = str(
+            f"host={db['host']} port={db['port']} dbname={db['dbname']} user={db['user']} password={db['password']}")
+        conn = psycopg2.connect(connStr)
+        cur = conn.cursor()
+
+        # Query it for column names
+        sql = str(f"""select column_name, data_type 
+                  from information_schema.columns
+                  where table_schema NOT IN ('information_schema', 'pg_catalog') and 
+                  table_name='{tableName}'""")
+        try:
+            cur.execute(sql)
+        except psycopg2.Error as e:
+            print(f"Error: getAnonTableCharacteristics() query: '{e}'")
+            self.cleanUp(cleanUpCache=False, doExit=True)
+        ans = cur.fetchall()
+
+        # Set up return dict
+        ret = {_row[0]: {'column_name': _row[0], 'column_type': _row[1]} for _row in ans}
+
+        conn.close()
+        return ret
+
     # Note that following is used internally, but we expose it to the
     # caller as well because it is a useful function for exploration
     def getColNamesAndTypes(self, dbType='rawDb', tableName=''):
@@ -1034,7 +1090,7 @@ class gdaAttack:
             if self._vb: print(f"   cache DB: {sql}")
             cur.execute(sql)
         sql = """CREATE TABLE IF NOT EXISTS tab
-                 (qid text, answer text)"""
+                 (qid text primary key, answer text)"""
         if self._vb: print(f"   cache DB: {sql}")
         cur.execute(sql)
         conn.close()
@@ -1261,6 +1317,9 @@ class gdaAttack:
             sys.exit("Error: Can't have more than 50 threads total")
 
     def _getCache(self, cur, query):
+        path = self._p['locCacheDir'] + "/" + self._p['name'] + ".db"
+        my_conn = sqlite3.connect(path, timeout=0.1)
+        my_cur = my_conn.cursor()
         # turn the query (dict) into a string
         qStr = self._dict2Str(query)
         if qStr is None:
@@ -1268,15 +1327,24 @@ class gdaAttack:
         sql = str(f"SELECT answer FROM tab where qid = '{qStr}'")
         if self._vb: print(f"   cache DB: {sql}")
         start = time.perf_counter()
-        try:
-            cur.execute(sql)
-        except sqlite3.Error as e:
-            print(f"getCache error '{e.args[0]}'")
+        for z in range(10):
+            try:
+                # cur.execute(sql)
+                my_cur.execute(sql)
+            except sqlite3.Error as e:
+                print(f"getCache error '{e.args[0]}' attempt: {z}")
+                continue
+            else:
+                break
+        else:
             return None
         end = time.perf_counter()
         self._op['numCacheGets'] += 1
         self._op['timeCacheGets'] += (end - start)
-        answer = cur.fetchone()
+        # answer = cur.fetchone() # frzmohammadali just to remember my stupidest bug ever
+        answer = my_cur.fetchone()
+        my_cur.close()
+        my_conn.close()
         if not answer:
             return None
         rtnDict = self._str2Dict(answer[0])
@@ -1284,6 +1352,8 @@ class gdaAttack:
 
     def _putCache(self, conn, cur, query, reply):
         # turn the query and reply (dict) into a string
+        # Establish connection to local cache
+        path = self._p['locCacheDir'] + "/" + self._p['name'] + ".db"
         qStr = self._dict2Str(query)
         if qStr is None:
             return
@@ -1296,20 +1366,39 @@ class gdaAttack:
         err = None
         for z in range(10):
             try:
-                cur.execute(sql)
-                conn.commit()
+                # cur.execute(sql)
+                # conn.commit()
+                my_conn = sqlite3.connect(path, timeout=0.1)
+                my_cur = my_conn.cursor()
+                my_cur.execute(sql)
+                my_conn.commit()
+
+            except sqlite3.IntegrityError as e:
+                if self._p['verbose'] or self._vb:
+                    print(f"putCache error [qid exists in cached queries] '{e.args[0]}' ")
+                break
             except sqlite3.OperationalError as e:
-                if self._p['verbose'] or self._vb:
-                    print(f"putCache error '{e.args[0]}'")
+                # database is locked
                 err = e
+                time.sleep(0.5)
                 continue
-            except sqlite3.Error as e:
+            except (sqlite3.Error, Exception) as e:
                 if self._p['verbose'] or self._vb:
-                    print(f"putCache error '{e.args[0]}'")
+                    print(f"putCache error '{e.args[0]}' attempt: {z}")
                 err = e
+                time.sleep(0.5)
                 continue
             else:
                 break
+            finally:
+                try:
+                    if my_cur:
+                        my_cur.close()
+                    if my_conn:
+                        my_conn.close()
+                except sqlite3.ProgrammingError:
+                    # cursor and connection is already closed
+                    pass
         else:
             # raise err
             if self._p['verbose'] or self._vb:
@@ -1579,7 +1668,7 @@ def signal_kill_handler(signum, frame):
     #              f' > sending termination signal to all. please wait ... ')
     cleanBgThreads()
     if atcObject:
-        atcObject.cleanUp()
+        atcObject.cleanUp(cleanUpCache=False)
 
     sys.exit(-1)
 
